/*
 * Copyright 1999 Finn Bock.
 * Permission to use, copy and distribute this software is hereby granted, 
 * provided that the above copyright notice appear in all copies and that 
 * both that copyright notice and this permission notice appear.
 * 
 * No Warranty
 * The software is provided "as is" without warranty of any kind.
 * 
 * If you have questions regarding this software, contact:
 *    Finn Bock, bckfnn@pipmail.dknet.dk
 * 
 * This program contains material copyrighted by:
 * Copyright 1991-1995 by Stichting Mathematisch Centrum, Amsterdam,
 * The Netherlands.
 *
 */

package jnios;

import java.io.*;
import java.util.*;
import org.python.core.*;

import java.lang.reflect.*;

import java.security.*;


import java.util.regex.*;



/**
 * <P>
 * This module provides access to operating system functionality that is
 * standardized by the C Standard and the POSIX standard (a thinly
 * disguised Unix interface).
 * 
 * <P>
 * The module is based on the original posixmodule.c except that all 
 * mistakes and errors are my own. Original author unknown.
 * 
 * <P>
 * <b>Do not import this module directly.</b>  Instead, import the
 * module <tt>jnios</tt>, which provides a <i>portable</i> version of this
 * interface.  On Unix, the <tt>jnios</tt> module provides a superset of
 * the <tt>posix</tt> interface.  On non-Unix operating systems the
 * <tt>posix</tt> module is not available, but a subset is always
 * available through the <tt>jnios</tt> interface.  Once <tt>jnios</tt> is
 * imported, there is <i>no</i> performance penalty in using it instead
 * of <tt>posix</tt>.  In addition, <tt>jnios</tt> provides some additional
 * functionality, such as automatically calling <tt>putenv()</tt>
 * when an entry in <tt>jnios.environ</tt> is changed.
 * 
 * <P>
 * The descriptions below are very terse; refer to the corresponding
 * Unix manual (or POSIX documentation) entry for more information.
 * Arguments called <i>path</i> refer to a pathname given as a string.
 * 
 * <P>
 * Errors are reported as exceptions; the usual exceptions are given for
 * type errors, while errors reported by the system calls raise
 * <tt>error</tt> (a synonym for the standard exception
 * <tt>OSError</tt>), described
 * below.
 * 
 * <P>
 * Module <tt>posix</tt> defines the following data items:
 * 
 * <P>
 * <dl><dt><b><tt>environ</tt></b>
 * <dd>
 * A dictionary or dictionary look-alike representing the string
 * environment at the time the interpreter was started.
 * For example,
 * <tt>posix.environ['HOME']</tt>
 * is the pathname of your home directory, equivalent to
 * <tt>getenv("HOME")</tt>
 * in C.
 * 
 * <P>
 * Modifying this dictionary does not affect the string environment
 * passed on by <tt>execv()</tt>, <tt>popen()</tt> or
 * <tt>system()</tt>; if you need to change the environment, pass
 * <tt>environ</tt> to <tt>execve()</tt> or add variable assignments and
 * export statements to the command string for <tt>system()</tt> or
 * <tt>popen()</tt>.
 * 
 * <P>
 * <i>However:</i> If you are using this module via the <tt>os</tt>
 * module (as you should - see the introduction above), <tt>environ</tt>
 * is a a mapping object that behaves almost like a dictionary but
 * invokes <tt>putenv()</tt> automatically whenever an item is changed.
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>error</tt></b>
 * <dd>
 * This exception is raised when a POSIX function returns a
 * POSIX-related error (e.g., not for illegal argument types).  The
 * accompanying value is a pair containing the numeric error code from
 * <tt>errno</tt> and the corresponding string, as would be printed by the
 * C function <tt>perror()</tt>.  See the module
 * <tt>errno</tt>, which contains names for the
 * error codes defined by the underlying operating system.
 * 
 * <P>
 * When exceptions are classes, this exception carries two attributes,
 * <tt>errno</tt> and <tt>strerror</tt>.  The first holds the value of
 * the C <tt>errno</tt> variable, and the latter holds the
 * corresponding error message from <tt>strerror()</tt>.  For
 * exceptions that involve a file system path (e.g. <tt>chdir</tt> or
 * <tt>unlink</tt>), the exception instance will contain a third attribute
 * <tt>filename</tt> which is the file name passed to the
 * function.
 * 
 * <P>
 * When exceptions are strings, the string for the exception is
 * <tt>'OSError'</tt>.
 * </dl>
 * 
 * <P>
 * It defines the following functions and constants:
 * 
 * <P>
 * <dl><dt><b><tt>chdir</tt></b> (<var>path</var>)
 * <dd>
 * Change the current working directory to <i>path</i>.
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>chmod</tt></b> (<var>path, mode</var>)
 * <dd>
 * Change the mode of <i>path</i> to the numeric <i>mode</i>.
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>chown</tt></b> (<var>path, uid, gid</var>)
 * <dd>
 * Change the owner and group id of <i>path</i> to the numeric <i>uid</i>
 * and <i>gid</i>.
 * (Not on MS-DOS.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>close</tt></b> (<var>fd</var>)
 * <dd>
 * Close file descriptor <i>fd</i>.
 * 
 * <P>
 * Note: this function is intended for low-level I/O and must be applied
 * to a file descriptor as returned by <tt>open()</tt> or
 * <tt>pipe()</tt>.  To close a ``file object'' returned by the
 * built-in function <tt>open()</tt> or by <tt>popen()</tt> or
 * <tt>fdopen()</tt>, use its <tt>close()</tt> method.
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>dup</tt></b> (<var>fd</var>)
 * <dd>
 * Return a duplicate of file descriptor <i>fd</i>.
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>dup2</tt></b> (<var>fd, fd2</var>)
 * <dd>
 * Duplicate file descriptor <i>fd</i> to <i>fd2</i>, closing the latter
 * first if necessary.
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>execv</tt></b> (<var>path, args</var>)
 * <dd>
 * Execute the executable <i>path</i> with argument list <i>args</i>,
 * replacing the current process (i.e., the Python interpreter).
 * The argument list may be a tuple or list of strings.
 * (Not on MS-DOS.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>execve</tt></b> (<var>path, args, env</var>)
 * <dd>
 * Execute the executable <i>path</i> with argument list <i>args</i>,
 * and environment <i>env</i>,
 * replacing the current process (i.e., the Python interpreter).
 * The argument list may be a tuple or list of strings.
 * The environment must be a dictionary mapping strings to strings.
 * (Not on MS-DOS.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>_exit</tt></b> (<var>n</var>)
 * <dd>
 * Exit to the system with status <i>n</i>, without calling cleanup
 * handlers, flushing stdio buffers, etc.
 * (Not on MS-DOS.)
 * 
 * <P>
 * Note: the standard way to exit is <tt>sys.exit(<i>n</i>)</tt>.
 * <tt>_exit()</tt> should normally only be used in the child process
 * after a <tt>fork()</tt>.
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>fork</tt></b> ()
 * <dd>
 * Fork a child process.  Return <tt>0</tt> in the child, the child's
 * process id in the parent.
 * (Not on MS-DOS.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>fstat</tt></b> (<var>fd</var>)
 * <dd>
 * Return status for file descriptor <i>fd</i>, like <tt>stat()</tt>.
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>ftruncate</tt></b> (<var>fd, length</var>)
 * <dd>
 * Truncate the file corresponding to file descriptor <i>fd</i>, 
 * so that it is at most <i>length</i> bytes in size.
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>getcwd</tt></b> ()
 * <dd>
 * Return a string representing the current working directory.
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>getegid</tt></b> ()
 * <dd>
 * Return the current process' effective group id.
 * (Not on MS-DOS.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>geteuid</tt></b> ()
 * <dd>
 * Return the current process' effective user id.
 * (Not on MS-DOS.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>getgid</tt></b> ()
 * <dd>
 * Return the current process' group id.
 * (Not on MS-DOS.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>getpgrp</tt></b> ()
 * <dd>
 * Return the current process group id.
 * (Not on MS-DOS.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>getpid</tt></b> ()
 * <dd>
 * Return the current process id.
 * (Not on MS-DOS.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>getppid</tt></b> ()
 * <dd>
 * Return the parent's process id.
 * (Not on MS-DOS.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>getuid</tt></b> ()
 * <dd>
 * Return the current process' user id.
 * (Not on MS-DOS.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>kill</tt></b> (<var>pid, sig</var>)
 * <dd>
 * Kill the process <i>pid</i> with signal <i>sig</i>.
 * (Not on MS-DOS.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>link</tt></b> (<var>src, dst</var>)
 * <dd>
 * Create a hard link pointing to <i>src</i> named <i>dst</i>.
 * (Not on MS-DOS.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>listdir</tt></b> (<var>path</var>)
 * <dd>
 * Return a list containing the names of the entries in the directory.
 * The list is in arbitrary order.  It does not include the special
 * entries <tt>'.'</tt> and <tt>'..'</tt> even if they are present in the
 * directory.
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>lseek</tt></b> (<var>fd, pos, how</var>)
 * <dd>
 * Set the current position of file descriptor <i>fd</i> to position
 * <i>pos</i>, modified by <i>how</i>: <tt>0</tt> to set the position
 * relative to the beginning of the file; <tt>1</tt> to set it relative to
 * the current position; <tt>2</tt> to set it relative to the end of the
 * file.
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>lstat</tt></b> (<var>path</var>)
 * <dd>
 * Like <tt>stat()</tt>, but do not follow symbolic links.  (On systems
 * without symbolic links, this is identical to <tt>stat()</tt>.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>mkfifo</tt></b> (<var>path</var><big>[</big><var>, mode</var><big>]</big>)
 * <dd>
 * Create a FIFO (a POSIX named pipe) named <i>path</i> with numeric mode
 * <i>mode</i>.  The default <i>mode</i> is <tt>0666</tt> (octal).  The current
 * umask value is first masked out from the mode.
 * (Not on MS-DOS.)
 * 
 * <P>
 * FIFOs are pipes that can be accessed like regular files.  FIFOs exist
 * until they are deleted (for example with <tt>os.unlink()</tt>).
 * Generally, FIFOs are used as rendezvous between ``client'' and
 * ``server'' type processes: the server opens the FIFO for reading, and
 * the client opens it for writing.  Note that <tt>mkfifo()</tt>
 * doesn't open the FIFO -- it just creates the rendezvous point.
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>mkdir</tt></b> (<var>path</var><big>[</big><var>, mode</var><big>]</big>)
 * <dd>
 * Create a directory named <i>path</i> with numeric mode <i>mode</i>.
 * The default <i>mode</i> is <tt>0777</tt> (octal).  On some systems,
 * <i>mode</i> is ignored.  Where it is used, the current umask value is
 * first masked out.
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>nice</tt></b> (<var>increment</var>)
 * <dd>
 * Add <i>increment</i> to the process' ``niceness''.  Return the new
 * niceness.  (Not on MS-DOS.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>open</tt></b> (<var>file, flags</var><big>[</big><var>, mode</var><big>]</big>)
 * <dd>
 * Open the file <i>file</i> and set various flags according to
 * <i>flags</i> and possibly its mode according to <i>mode</i>.
 * The default <i>mode</i> is <tt>0777</tt> (octal), and the current umask
 * value is first masked out.  Return the file descriptor for the newly
 * opened file.
 * 
 * <P>
 * For a description of the flag and mode values, see the Unix or C
 * run-time documentation; flag constants (like <tt>O_RDONLY</tt> and
 * <tt>O_WRONLY</tt>) are defined in this module too (see below).
 * 
 * <P>
 * Note: this function is intended for low-level I/O.  For normal usage,
 * use the built-in function <tt>open()</tt>, which returns a ``file
 * object'' with <tt>read()</tt> and <tt>write()</tt> methods (and many
 * more).
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>pipe</tt></b> ()
 * <dd>
 * Create a pipe.  Return a pair of file descriptors <tt>(<i>r</i>,
 * <i>w</i>)</tt> usable for reading and writing, respectively.
 * (Not on MS-DOS.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>plock</tt></b> (<var>op</var>)
 * <dd>
 * Lock program segments into memory.  The value of <i>op</i>
 * (defined in <tt>&lt;sys/lock.h&gt;</tt>) determines which segments are locked.
 * (Not on MS-DOS.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>popen</tt></b> (<var>command</var><big>[</big><var>, mode</var><big>[</big><var>, bufsize</var><big>]</big><big>]</big>)
 * <dd>
 * Open a pipe to or from <i>command</i>.  The return value is an open
 * file object connected to the pipe, which can be read or written
 * depending on whether <i>mode</i> is <tt>'r'</tt> (default) or <tt>'w'</tt>.
 * The <i>bufsize</i> argument has the same meaning as the corresponding
 * argument to the built-in <tt>open()</tt> function.  The exit status of
 * the command (encoded in the format specified for <tt>wait()</tt>) is
 * available as the return value of the <tt>close()</tt> method of the file
 * object, except that when the exit status is zero (termination without
 * errors), <tt>None</tt> is returned.
 * (Not on MS-DOS.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>putenv</tt></b> (<var>varname, value</var>)
 * <dd>
 * Set the environment variable named <i>varname</i> to the string
 * <i>value</i>.  Such changes to the environment affect subprocesses
 * started with <tt>os.system()</tt>, <tt>os.popen()</tt> or
 * <tt>os.fork()</tt> and <tt>os.execv()</tt>.  (Not on all systems.)
 * 
 * <P>
 * When <tt>putenv()</tt> is
 * supported, assignments to items in <tt>os.environ</tt> are automatically
 * translated into corresponding calls to <tt>putenv()</tt>; however,
 * calls to <tt>putenv()</tt> don't update <tt>os.environ</tt>, so it is
 * actually preferable to assign to items of <tt>os.environ</tt>.  
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>strerror</tt></b> (<var>code</var>)
 * <dd>
 * Return the error message corresponding to the error code in <i>code</i>.
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>read</tt></b> (<var>fd, n</var>)
 * <dd>
 * Read at most <i>n</i> bytes from file descriptor <i>fd</i>.
 * Return a string containing the bytes read.
 * 
 * <P>
 * Note: this function is intended for low-level I/O and must be applied
 * to a file descriptor as returned by <tt>open()</tt> or
 * <tt>pipe()</tt>.  To read a ``file object'' returned by the
 * built-in function <tt>open()</tt> or by <tt>popen()</tt> or
 * <tt>fdopen()</tt>, or <tt>sys.stdin</tt>, use its
 * <tt>read()</tt> or <tt>readline()</tt> methods.
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>readlink</tt></b> (<var>path</var>)
 * <dd>
 * Return a string representing the path to which the symbolic link
 * points.  (On systems without symbolic links, this always raises
 * <tt>error</tt>.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>remove</tt></b> (<var>path</var>)
 * <dd>
 * Remove the file <i>path</i>.  See <tt>rmdir()</tt> below to remove a
 * directory.  This is identical to the <tt>unlink()</tt> function
 * documented below.
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>rename</tt></b> (<var>src, dst</var>)
 * <dd>
 * Rename the file or directory <i>src</i> to <i>dst</i>.
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>rmdir</tt></b> (<var>path</var>)
 * <dd>
 * Remove the directory <i>path</i>.
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>setgid</tt></b> (<var>gid</var>)
 * <dd>
 * Set the current process' group id.
 * (Not on MS-DOS.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>setpgrp</tt></b> ()
 * <dd>
 * Calls the system call <tt>setpgrp()</tt> or <tt>setpgrp(0,
 * 0)</tt> depending on which version is implemented (if any).  See the
 * Unix manual for the semantics.
 * (Not on MS-DOS.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>setpgid</tt></b> (<var>pid, pgrp</var>)
 * <dd>
 * Calls the system call <tt>setpgid()</tt>.  See the Unix manual
 * for the semantics.
 * (Not on MS-DOS.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>setsid</tt></b> ()
 * <dd>
 * Calls the system call <tt>setsid()</tt>.  See the Unix manual
 * for the semantics.
 * (Not on MS-DOS.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>setuid</tt></b> (<var>uid</var>)
 * <dd>
 * Set the current process' user id.
 * (Not on MS-DOS.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>stat</tt></b> (<var>path</var>)
 * <dd>
 * Perform a <tt>stat()</tt> system call on the given path.  The
 * return value is a tuple of at least 10 integers giving the most
 * important (and portable) members of the <i>stat</i> structure, in the
 * order
 * <tt>st_mode</tt>,
 * <tt>st_ino</tt>,
 * <tt>st_dev</tt>,
 * <tt>st_nlink</tt>,
 * <tt>st_uid</tt>,
 * <tt>st_gid</tt>,
 * <tt>st_size</tt>,
 * <tt>st_atime</tt>,
 * <tt>st_mtime</tt>,
 * <tt>st_ctime</tt>.
 * More items may be added at the end by some implementations.
 * (On MS-DOS, some items are filled with dummy values.)
 * 
 * <P>
 * Note: The standard module <tt>stat</tt> defines
 * functions and constants that are useful for extracting information
 * from a stat structure.
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>symlink</tt></b> (<var>src, dst</var>)
 * <dd>
 * Create a symbolic link pointing to <i>src</i> named <i>dst</i>.  (On
 * systems without symbolic links, this always raises <tt>error</tt>.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>system</tt></b> (<var>command</var>)
 * <dd>
 * Execute the command (a string) in a subshell.  This is implemented by
 * calling the Standard C function <tt>system()</tt>, and has the
 * same limitations.  Changes to <tt>posix.environ</tt>, <tt>sys.stdin</tt>
 * etc. are not reflected in the environment of the executed command.
 * The return value is the exit status of the process encoded in the
 * format specified for <tt>wait()</tt>.
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>tcgetpgrp</tt></b> (<var>fd</var>)
 * <dd>
 * Return the process group associated with the terminal given by
 * <i>fd</i> (an open file descriptor as returned by <tt>open()</tt>).
 * (Not on MS-DOS.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>tcsetpgrp</tt></b> (<var>fd, pg</var>)
 * <dd>
 * Set the process group associated with the terminal given by
 * <i>fd</i> (an open file descriptor as returned by <tt>open()</tt>)
 * to <i>pg</i>.
 * (Not on MS-DOS.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>times</tt></b> ()
 * <dd>
 * Return a 5-tuple of floating point numbers indicating accumulated (CPU
 * or other)
 * times, in seconds.  The items are: user time, system time, children's
 * user time, children's system time, and elapsed real time since a fixed
 * point in the past, in that order.  See the Unix
 * manual page <i>times</i>(2).  (Not on MS-DOS.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>umask</tt></b> (<var>mask</var>)
 * <dd>
 * Set the current numeric umask and returns the previous umask.
 * (Not on MS-DOS.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>uname</tt></b> ()
 * <dd>
 * Return a 5-tuple containing information identifying the current
 * operating system.  The tuple contains 5 strings:
 * <tt>(<i>sysname</i>, <i>nodename</i>, <i>release</i>, <i>version</i>,
 * <i>machine</i>)</tt>.  Some systems truncate the nodename to 8
 * characters or to the leading component; a better way to get the
 * hostname is <tt>socket.gethostname()</tt>287
 * 
 * or even
 * <tt>socket.gethostbyaddr(socket.gethostname())</tt>288.
 * (Not on MS-DOS, nor on older Unix systems.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>unlink</tt></b> (<var>path</var>)
 * <dd>
 * Remove the file <i>path</i>.  This is the same function as <tt>remove</tt>;
 * the <tt>unlink</tt> name is its traditional Unix name.
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>utime</tt></b> (<var>path, (atime, mtime)</var>)
 * <dd>
 * Set the access and modified time of the file to the given values.
 * (The second argument is a tuple of two items.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>wait</tt></b> ()
 * <dd>
 * Wait for completion of a child process, and return a tuple containing
 * its pid and exit status indication: a 16-bit number, whose low byte is
 * the signal number that killed the process, and whose high byte is the
 * exit status (if the signal number is zero); the high bit of the low
 * byte is set if a core file was produced.  (Not on MS-DOS.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>waitpid</tt></b> (<var>pid, options</var>)
 * <dd>
 * Wait for completion of a child process given by proces id, and return
 * a tuple containing its pid and exit status indication (encoded as for
 * <tt>wait()</tt>).  The semantics of the call are affected by the
 * value of the integer <i>options</i>, which should be <tt>0</tt> for
 * normal operation.  (If the system does not support
 * <tt>waitpid()</tt>, this always raises <tt>error</tt>.  Not on
 * MS-DOS.)
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>write</tt></b> (<var>fd, str</var>)
 * <dd>
 * Write the string <i>str</i> to file descriptor <i>fd</i>.
 * Return the number of bytes actually written.
 * 
 * <P>
 * Note: this function is intended for low-level I/O and must be applied
 * to a file descriptor as returned by <tt>open()</tt> or
 * <tt>pipe()</tt>.  To write a ``file object'' returned by the
 * built-in function <tt>open()</tt> or by <tt>popen()</tt> or
 * <tt>fdopen()</tt>, or <tt>sys.stdout</tt> or <tt>sys.stderr</tt>, use
 * its <tt>write()</tt> method.
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>WNOHANG</tt></b>
 * <dd>
 * The option for <tt>waitpid()</tt> to avoid hanging if no child
 * process status is available immediately.
 * </dl>
 * 
 * <P>
 * <dl><dt><b><tt>O_RDONLY</tt></b>
 * <dd>
 * <dt><b><tt>O_WRONLY</tt></b><dd>
 * <dt><b><tt>O_RDWR</tt></b><dd>
 * <dt><b><tt>O_NDELAY</tt></b><dd>
 * <dt><b><tt>O_NONBLOCK</tt></b><dd>
 * <dt><b><tt>O_APPEND</tt></b><dd>
 * <dt><b><tt>O_DSYNC</tt></b><dd>
 * <dt><b><tt>O_RSYNC</tt></b><dd>
 * <dt><b><tt>O_SYNC</tt></b><dd>
 * <dt><b><tt>O_NOCTTY</tt></b><dd>
 * <dt><b><tt>O_CREAT</tt></b><dd>
 * <dt><b><tt>O_EXCL</tt></b><dd>
 * <dt><b><tt>O_TRUNC</tt></b><dd>
 * Options for the <tt>flag</tt> argument to the <tt>open()</tt> function.
 * These can be bit-wise OR'd together.
 * </dl>
 *
 * @author Finn Bock, bckfnn@pipmail.dknet.dk
 * @version $Id: posix.java,v 1.1 1999/10/12 17:15:38 fb Exp $
*/


public class posix implements InitModule {

    public void initModule(PyObject dict) {
	// A hack to check for a JPython-1.0.3 environment.
	// Unfortunatly, it means this will not compile with 1.0.3; oh well.
	try {
	    error = Py.OSError;
	} catch (java.lang.NoSuchFieldError e) {
	    error = new PyString("os.error");
	}

	dict.__delitem__("initModule");

	Hashtable names = new Hashtable();
	PyObject keys = dict.invoke("keys");
	for (int i = 0; i < keys.__len__(); i++) {
	    PyObject key = keys.__getitem__(i);
	    PyObject value = dict.__getitem__(key);
	    names.put(key.toString(), value);
	}

	dict.invoke("clear");

	init(dict, error);

	String[] nativeNames = getNames();
	for (int i = 0; i < nativeNames.length; i++) {
	    dict.__setitem__(nativeNames[i].intern(), (PyObject)names.get(nativeNames[i]));
	}

	PyObject environ = new PyDictionary();

	Hashtable env = convertenviron();
	for (Enumeration e = env.keys(); e.hasMoreElements(); ) {
	    String key = (String) e.nextElement();
	    environ.__setitem__(key, new PyString((String)env.get(key)));
	    //println(key + " = " + env.get(key));
	}

	dict.__setitem__("environ", environ);
    }

    private final static native String[] getNames();


	static
	{


		// System.out.println("posix.java static{} System.loadLibrary call()");
		// println()lCeBuR[hprintf()ɒu邱Ƃ͂łȂ(Cũ[hO)

		//System.loadLibrary("posix");	//Avbgp



		
		AccessController.doPrivileged(new PrivilegedAction() {	//AvbgΉ
			public Object run() {
				try {
					System.loadLibrary("posix");
				}catch(Exception e) {
		    			e.printStackTrace();
					
	  			}
					return null;
	    		}
		});
		

		// System.out.println("posix.java static{} System.loadLibrary returned");



	}





    private final native void init(PyObject dict, PyObject error);

    private final static native Hashtable convertenviron();


    public static PyObject error;


    /**
     * chdir(path) -> None.
     * Change the current working directory to the specified path.
     */
    public final static native int chdir(String dir);


    /**
     * chmod(path, mode) -> None.
     * Change the access permissions of a file.
     */
    public final static native int chmod(String dir, int mode);


    /**
     * fsync(fildes) -> None.
     * force write of file with filedescriptor to disk.
     */
    public final static native void fsync(int fd);


    /**
     * fdatasync(fildes) -> None.
     * force write of file with filedescriptor to disk.
     * does not force update of metadata.
     */
    public final static native void fdatasync(int fd);


    /**
     * chown(path, uid, gid) -> None.
     * Change the owner and group id of path to the numeric uid and gid.
     */
    public final static native void chown(String dir, int uid, int gid);


    /**
     * getcwd() -> path.
     * Return a string representing the current working directory.
     */
    public final static native String getcwd();

    /**
     * link(src, dst) -> None.
     * Create a hard link to a file.
     */
    public final static native int link(String src, String dst);
	/*
	߂lvoidint֕ύX
	*/
		

    /**
     * listdir(path) -> list_of_strings.
     * Return a list containing the names of the entries in the directory.
     * <p>
     * 	path: path of directory to list.
     * <p>
     * The list is in arbitrary order.  It does not include the special
     * entries '.' and '..' even if they are present in the directory.
     */
    public final static PyList listdir(String path) {
	
	//return __builtin__.list(Py.java2py(new File(path).list()));
    //NKύX
   	
   	return new PyList( Py.java2py(new File(path).list()) );
   
    
    }


    /**
     * lstat(path) -> (mode,ino,dev,nlink,uid,gid,size,atime,mtime,ctime).
     * Like stat(path), but do not follow symbolic links.
     */
    public final static native PyTuple lstat(String file);

 

    /**
     * mkdir(path [, mode=0777]) -> None.
     * Create a directory.
     */
    public final static void mkdir(String path) {
	mkdir(path, 0777);
    }


    /**
     * mkdir(path, mode=0777) -> None.
     * Create a directory.
     */
    public final static native int mkdir(String dir, int mode);


    /**
     * nice(inc) -> new_priority.
     * Decrease the priority of process and return new priority.
     */
    public final static native int nice(int incr);


    /**
     * readlink(path) -> path.
     * Return a string representing the path to which the symbolic link points.
     */
    public final static native String readlink(String file);


    /**
     * rename(old, new) -> None.
     * Rename a file or directory.
     */
    public final static native int rename(String src, String dst);


    /**
     * rmdir(path) -> None.
     * Remove a directory.
     */
    public final static native int rmdir(String dir);


    /**
     * stat(path) -> (mode,ino,dev,nlink,uid,gid,size,atime,mtime,ctime).
     * Perform a stat system call on the given path.
     */
    public final static native PyTuple stat(String file);


    /**
     * symlink(src, dst) -> None.
     * Create a symbolic link.
     */
    public final static native void symlink(String src, String dst);


    /**
     * access(path, mode) -> 1 if granted, 0 otherwise.
     * Test for access to a file.
     */
    public final static native long access(String path, int mode);



    /**
     * ttyname(fd) -> String.
     * Return the name of the terminal device connected to 'fd'.
     */
    public final static native String ttyname(int fd);

    /**
     * system(command) -> exit_status.
     * Execute the command (a string) in a subshell.
     */
    public final static native int system(String cmd);


    /**
     * umask(new_mask) -> old_mask.
     * Set the current numeric umask and return the previous umask.
     */
    public final static native int umask(int mask);


    /**
     * uname() -> (sysname, nodename, release, version, machine).
     * Return a tuple identifying the current operating system.
     */
    
    /* 08/02/04 N.K delete
    public final static native PyTuple uname();
	*/
	

    /**
     * unlink(path) -> None.
     * Remove a file (same as remove(path)).
     */
    public final static native int unlink(String file);
	/*
	߂lvoidintɕύX
	*/

    /**
     * remove(path) -> None.
     * Remove a file (same as unlink(path)).
     */
    public final static void remove(String file) {
	unlink(file);
    }


    /**
     * utime(path, (atime, utime)) -> None.
     * Set the access and modified time of the file to the given values.
     */
    public final static native int utime(String file, int[] times);


    /**
     * times() -> (utime, stime, cutime, cstime, elapsed_time).
     * Return a tuple of floating point numbers indicating process times.
     */
    public final static native long[] times();


    /**
     * _exit(status).
     * Exit to the system with specified status, without normal exit processing.
     */
    public final static native void _exit(int status);

    /**
     * execv(path, args).
     * Execute an executable path with arguments, replacing current process.
     * <p>
     * path: path of executable file<br>
     * args: tuple or list of strings<br>
     */
    public final static native void execv(String path, String[] args);



    /**
     * execve(path, args, env).
     * Execute a path with arguments and environment, replacing current process.
     * <p>
     * path: path of executable file<br>
     * args: tuple or list of arguments<br>
     * env: dictonary of strings mapping to strings<br>
     */
    public final static void execve(String path, String[] args, PyObject env) {
	PyObject envKeys = env.invoke("keys");
	String[] envArr = new String[envKeys.__len__()];
	for (int i = 0; i < envKeys.__len__(); i++) {
	    PyObject key = envKeys.__getitem__(i);
	    PyObject value = env.__getitem__(key);
	    if (getName().equals("os2")) { 
	        /* Omit Pseudo-Env Vars that Would Confuse Programs if Passed On */
	        if (key.equals("BEGINLIBPATH") || key.equals("ENDLIBPATH"))
		    continue;
	    }
	    envArr[i] = key.toString() + "=" + value.toString();
	}

	execve0(path, args, envArr);
    }


    public final static native int execve0(String path, String[] args, String[] env);



    /**
     * spawnv(mode, path, args).
     * Execute an executable path with arguments, replacing current process.
     * <p>
     * mode: mode of process creation<br>
     * path: path of executable file<br>
     * args: tuple or list of strings
     */
    public final static native int spawnv(int mode, String path, String[] args);


    /**
     * spawnve(mode, path, args, env).
     * Execute a path with arguments and environment, replacing current process.
     * <p>
     * mode: mode of process creation<br>
     * path: path of executable file<br>
     * args: tuple or list of arguments<br>
     * env: dictonary of strings mapping to strings<br>
     */
    public final static int spawnve(int mode, String path, String[] args, PyObject env) {
	PyObject envKeys = env.invoke("keys");
	String[] envArr = new String[envKeys.__len__()];
	for (int i = 0; i < envKeys.__len__(); i++) {
	    PyObject key = envKeys.__getitem__(i);
	    PyObject value = env.__getitem__(key);
	    envArr[i] = key.toString() + "=" + value.toString();
	}

	return spawnve0(mode, path, args, envArr);
    }

    private final static native int spawnve0(int mode, String path, String[] args, String[] env);



    /**
     * fork() -> pid.
     * Fork a child process.
     * <p>
     * Return 0 to child process and PID of child to parent process.
     */
    public final static native int fork();


    /**
     * getegid() -> egid.
     * Return the current process's effective group id.
     */
    public final static native int getegid();


    /**
     * geteuid() -> euid.
     * Return the current process's effective user id.
     */
    public final static native int geteuid();


    /**
     * getgid() -> gid.
     * Return the current process's group id.
     */
    public final static native int getgid();


    /**
     * getpid() -> pid.
     * Return the current process id
     */
    public final static native int getpid();


    /**
     * getpgrp() -> pgrp.
     * Return the current process group id.
     */
    public final static native int getpgrp();


    /**
     * getppid() -> ppid.
     * Return the parent's process id.
     */
    public final static native int getppid();


    /**
     * getuid() -> uid.
     * Return the current process's user id.
     */
    public final static native int getuid();


    /**
     * kill(pid, sig) -> None.
     * Kill a process with a signal.
     */
    public final static native int kill(int pid, int sig);


    /**
     * plock(op) -> None.
     * Lock program segments into memory.
     */
    public final static native void plock(int op);


    /**
     * popen(command [, mode='r' [, bufsize]]) -> pipe.
     * Open a pipe to/from a command returning a file object.
     */
    public final static PosixFile popen(String name) {
	return popen(name, "r", -1);
    }


    /**
     * popen(command [, mode='r' [, bufsize]]) -> pipe.
     * Open a pipe to/from a command returning a file object.
     */
    public final static PosixFile popen(String name, String mode) {
	return popen(name, mode, -1);
    }


    /**
     * popen(command [, mode='r' [, bufsize]]) -> pipe.
     * Open a pipe to/from a command returning a file object.
     */
    public final static PosixFile popen(String name, String mode, int bufsize) {
	PosixFile f = new PosixFile(popen0(name, mode, bufsize), name, mode, pclose0());
	f.setbufsize(bufsize);
	return f;
    }


    private final static native long popen0(String name, String mode, int bufsize);

    private final static native long pclose0();



    /**
     * setuid(uid) -> None.
     * Set the current process's user id.
     */
    public final static native int setuid(int uid);


    /**
     * setgid(gid) -> None.
     * Set the current process's group id.
     */
    public final static native int setgid(int gid);


    /**
     * setpgrp() -> None.
     * Make this process a session leader.
     */
    public final static native void setpgrp();


    /**
     * wait() -> (pid, status).
     * Wait for completion of a child process.
     */
    public final static native PyTuple wait(int pid, int options);


    /**
     * waitpid(pid, options) -> (pid, status).
     * Wait for completion of a give child process.
     */
    public final static native PyTuple waitpid(int pid, int options);


    /**
     * setsid() -> None.
     * Call the system call setsid().
     */
    public final static native void setsid();


    /**
     * setpgid(pid, pgrp) -> None.
     * Call the system call setpgid().
     */
    public final static native void setpgid();


    /**
     * tcgetpgrp(fd) -> pgid.
     * Return the process group associated with the terminal given by a fd.
     */
    public final static native int tcgetpgrp(int fd);


    /**
     * tcsetpgrp(fd, pgid) -> None.
     * Set the process group associated with the terminal given by a fd.
     */
    public final static native void tcsetpgrp(int fd, int pgid);



    /**
     * open(filename, flag mode=0777) -> fd.
     * Open a file (for low level IO).
     */
    public final static native int open(String file, int flag, int mode);


    /**
     * open(filename, flag) -> fd.
     * Open a file (for low level IO).
     */
    public final static int open(String file, int flag) {
	return open(file, flag, 0777);
    }


    /**
     * close(fd) -> None.
     * Close a file descriptor (for low level IO).
     */
    public final static native long close(int fd);


    /**
     * dup(fd) -> fd2-
     * Return a duplicate of a file descriptor.
     */
    public final static native int dup(int fd);


    /**
     * dup2(fd, fd2) -> None.
     * Duplicate file descriptor.
     */
    public final static native long dup2(int fd, int fd2);
	//09/01/09ύX.

    /**
     * lseek(fd, pos, how) -> newpos.
     * Set the current position of a file descriptor.
     */
    public final static native long lseek(int fd, long pos, int how);


    /**
     * read(fd, buffersize) -> string.
     * Read a file descriptor.
     */

	//08/06/09 Stringx[X͔̂̂p~
    //public final static native String read(int fd, int size);
	//08/06/09 Stringx[X͔̂̂p~

	public final static native int	read(int fd, byte[] buf);



    /**
     * write(fd, string) -> byteswritten.
     * Write a string to a file descriptor.
     */
    public final static native int write(int fd, byte[] buf);


    /**
     * fstat(fd) -> (mode, ino, dev, nlink, uid, gid, size, atime, mtime, ctime).
     * Like stat(), but for an open file descriptor.
     */
    public final static native PyTuple fstat(int fd);



    /**
     * fdopen(fd, [, mode='r' [, bufsize]]) -> file_object.
     * Return an open file object connected to a file descriptor.";
     */
    public final static PosixFile fdopen(int fd) {
	return fdopen(fd, "r", -1);
    }

    /**
     * fdopen(fd, [, mode='r' [, bufsize]]) -> file_object.
     * Return an open file object connected to a file descriptor.";
     */
    public final static PosixFile fdopen(int fd, String mode) {
	return fdopen(fd, mode, -1);
    }

    /**
     * fdopen(fd, [, mode='r' [, bufsize]]) -> file_object.
     * Return an open file object connected to a file descriptor.";
     */
    public final static PosixFile fdopen(int fd, String mode, int bufsize) {
	PosixFile f = new PosixFile(fdopen0(fd, mode), "(fdopen)", mode, fclose0());
	f.setbufsize(bufsize);
	return f;
    }


    private final static native long fdopen0(int fd, String mode);

    private final static native long fclose0();

    /**
     * pipe() -> (read_end, write_end).
     * Create a pipe.
     */
    //public final static native PyTuple pipe();
	//ύXς

    /**
     * mkfifo(file, [, mode=0666]) -> None.
     * Create a FIFO (a POSIX named pipe).
     */
    public final static native void mkfifo(String file, int mode);


    /**
     * ftruncate(fd, length) -> None.
     * Truncate a file to a specified length.
     */
    public final static native void ftruncate(int fd, int len);


    /**
     * putenv(key, value) -> None.
     * Change or add an environment variable.
     */
    public final static native void putenv(String key, String value);


    /**
     * strerror(code) -> string.
     * Translate an error code to a message string.
     */
    public final static native String strerror(int code);


    /**
     * WIFSTOPPED(status) -> Boolean
     * Return true if the process returning 'status' was stopped.
     */
    public final static native int WIFSTOPPED(int status);


    /**
     * WIFSIGNALED(status) -> Boolean.
     * Return true if the process returning 'status' was terminated by a signal.
     */
    public final static native int WIFSIGNALED(int status);


    /**
     * WIFEXITED(status) -> Boolean.
     * Return true if the process returning 'status' exited using the exit().
     */
    public final static native int WIFEXITED(int status);


    /**
     * WEXITSTATUS(status) -> integer.
     * Return the process return code from 'status'.
     */
    public final static native int WEXITSTATUS(int status);


    /**
     * WTERMSIG(status) -> integer.
     * Return the signal that terminated the process that provided the 'status' value.
     */
    public final static native int WTERMSIG(int status);


    /**
     * WSTOPSIG(status) -> integer.
     * Return the signal that stopped the process that provided the 'status' value.
     */
    public final static native int WSTOPSIG(int status);


    /**
     * fstatvfs(fd) -> 
     * (bsize, frsize, blocks, bfree, bavail, files, ffree, favail, flag, namemax).
     * Perform an fstatvfs system call on the given fd.
     */
    public final static native PyTuple fstatvfs(int fd);


    /**
     * statvfs(path) -> 
     * (bsize, frsize, blocks, bfree, bavail, files, ffree, favail, flag, namemax).
     * Perform a statvfs system call on the given path.
     */
    public final static native PyTuple fstatvfs(String path);



    public static void main(String[] args) {
	String[] names = getNames();
	for (int i = 0; i < names.length; i++)
	    println(names[i]);
    }

    public final static native String getName();
    
    
    
    
    
    
    
    
    
    
/*************************************07/08/25 Naoya Kawaguchi Append*****************************************************/

    public final static native int creat(String path, int mode);
    
	public final static native int time(long[] time_t);
	
	public final static native int mknod(String path, int mode, int dev);
	
	public final static native int lchown16(String path, int user, int group);
	
	public final static native int stime (long[] p_time_t);
	
	public final static native int ptrace (int request, int pid, byte[] address, byte[] data);
	
	public final static native int alarm ( int seconds );
	
	public final static native int pause();
	
	public final static native void sync();
    
	public final static native int pipe( int[] pipes);

	public final static native void hogehoge();
	

	
	private static String printIOName = "stderr";	// posix.print()ŎgpIO̖O. posix.setPrintFD()ŕύX\.

	public final static synchronized long setPrintFILE(String name)
	{
		//posix.print()ŎgpFDݒ肷.
		//w\Ȃ̂ stdin, stdout, stderr
		// ƃlCeBuR[hfp̒lԂAs-1Ԃ.

		long fp_val = getStdIOFILE( name );
		if ( fp_val >= 0 )
		{
			printIOName = new String( name );
		}
		return fp_val;
	}

	public final static void println( String str )
	{
		print( str + "\n" );
	}

	public final static void print(String str)
	{
		long fp_val = getStdIOFILE( printIOName );
		print( fp_val, str );
	}

	public final static native long getStdIOFILE(String fd_name);
	// Wo͂Ŏgp stdin, stdout, stderrȂǂ̃t@C|C^𓾂.lCeBuR[hŔFłȂ萔l̖Ow肷-1Ԃ.

	public final static native void print( long fp, String str);
	//fobOp̕Wo͋@\.o͐̓ftHgł͕Wo͂AsetPrintFD()ŕύXł.
	//	fpɂ̓lCeBuR[h̃t@C|C^̒lw肷.

	public final static native String gets();
	public final static native String[] getenvs();



/************************************ 07/12/29 Socket API append ****************************************************************/


/*
	long[] addrArr struct sockaddr_in\̂̑
	[0] = addrlen(lbg[NoCgI[_[ł邱Ƃɒ)
	[1] = sin_family;
	[2] = sin_addr.s_addr;
	[3] = sin_port;
	Ƃ

AF_UNIX 1
AF_INET 2
AF_AX25 3
AF_NETROM 6
AF_BRIDGE 7
AF_BRIDGE 7
AF_MAX 12



*/

	

	public final static native int socket(int family, int type, int protocol);

	public final static int socket(String family, String type, String protocol)
	{	//socket() ̃bp[֐. "AF_INET"̂悤Ɏwł.

		int fd = socket(	(int)getSocketConstValue( family),
					(int)getSocketConstValue( type ),
					(int)getSocketConstValue( protocol)
				);

		return fd;
	}


	public final static native int bind(int fd, long[] addrArr);
	
	public final static int bind(int fd, String family, String address, String port)
	{
		//bind()̃bp[֐. 𕶎œn
		long[] addrArr = getAddrArr( family, address, port);

		return bind(fd, addrArr);
	}



	public final static native int connect(int fd, long[] addrArr);

	public final static int connect(int fd, String family, String address, String port)
	{
		//connect() ̃bp[֐. 𕶎œn

		long[] addrArr = getAddrArr( family, address, port);

		return connect(fd, addrArr);
	}

	public final static native int listen(int fd, int backlog);

	public final static native int accept(int fd, long[] addrArr);

	public final static native long getsockname(int fd, long[] addrArr);

	public final static native long getpeername(int fd, long[] addrArr);

	public final static native long setsockopt(int fd, int level, int optname, byte[] optval, int optlen );
	
	public final static native long getsockopt(int fd, int level, int optname, byte[] optval, int[] optlen );

	//public final static native long socketpair();


	public final static native int send(int fd, byte[] buf, int size, long flag);
	public final static native int recv(int fd, byte[] buf, int size, long flag);

	public final static native int sendto(int fd, byte[] buf, int size, long flag, long[] addrArr);

	public final static native int recvfrom(int fd, byte[] buf, int size, long flag, long[] addrArr);




	//ȉPosix-APIł͂ȂA̓s
	public final static native long inetaddr(String addrstr);
	public final static native String inetntoa(long addrlong);
	public final static native long htons(long host_endian_port);
	public final static native long getSocketConstValue(String valueName); 
		//Ǝ֐.AF_INETSOCK_STREAM,IPPROT_TCP Ȃǂ̒萔lԂ

	public final static long[] getAddrArr(long family, long addr, long port, long addrlen)
	{
		// Ǝ֐. addr_in\̂̑ւƂȂlongzԂ.
		// portɂ̓lbg[NoCgI[_[w
		//@܂addrlenKlbg[NoCgI[_[̐lw肷邱.

		long[] addrArr = new long[4];
		
		addrArr[0] = addrlen;
		addrArr[1] = family;
		addrArr[2] = addr;
		addrArr[3] = port;

		return addrArr;
	}
	public final static long[] getAddrArr(String family, String address, String port)
	{
		//Ǝ֐. addr_in\̂̑ւƂȂlongzԂ.
		//port́AzXg̃GfBAɊւ炸A{̒ltoString()
		long[] addrArr = getAddrArr(
						getSocketConstValue(family),
						(long)inetaddr( address),
						(long)htons( Long.parseLong(port) ),
						0 );
		return addrArr;
	}


	public final static long[] getAddrArr6( long sin6_family, long[] sin6_addr, long sin6_port, long sin6_flowinfo, long sin6_scope_id, long addrlen)
	{
		//Ǝ֐. sockaddr_in6\̂̑ւȂlongzԂ. Ȃsin6_addr128rbgȂ̂ŁA64rbgɕlong[]nB
		//̊֐ł,sin6_addr\long[]͂߂Ƃϐɑ΂AGfBAϊ͍sȂ.
		//ȂAŌ̗vfaddrleni[.Linuxł͎gp邱Ƃ͂Ȃ낤AÔ߁B

		/*
		struct sockaddr_in6
		{                                            
			sa_family_t       sin6_family;   //  AF_INET6              
			in_port_t         sin6_port;     //  port number           
			uint32_t          sin6_flowinfo; //  IPv6 flow infomation  
			struct in6_addr   sin6_addr;     //  IPv6 address          
			uint32_t          sin6_scope_id; //  scope id              
		};
		*/

		long[] addrArr = new long[ 7 ];


		addrArr[0]	= addrlen;
		addrArr[1]	= sin6_family;
		addrArr[2]	= sin6_port;
		addrArr[3]	= sin6_flowinfo;
		addrArr[4]	= sin6_addr[0];
		addrArr[5]	= sin6_addr[1];
		addrArr[6]	= sin6_scope_id;

		return addrArr;
	}


	public final static long[] getAddrArrUN( long sun_family, String sun_path, long addrlen)
	{
		//Ǝ֐. sockaddr_un\̂̑ւȂlongzԂ. Ȃsun_path108oCgchar^ł邪A1oCglong[]̗vfƂĊi[.
		//̊֐ł,sin6_addr\long[]͂߂Ƃϐɑ΂AGfBAϊ͍sȂ.
		//ȂAŌ̗vfaddrleni[.Linuxł͎gp邱Ƃ͂Ȃ낤AÔ߁B

		/*
		struct sockaddr_un
		{
			__SOCKADDR_COMMON(sun_); //= (Fedora8ł)sa_family_t݂̂錾}N
			char	sun_path[108];
		};
		*/

		long[] addrArr = new long[ 110 ];

		addrArr[0]	= addrlen;
		addrArr[1]	= sun_family;

		byte[]	sun_path_bytes = StringToBytes( sun_path );

		int sun_path_index = 2;
		for( int i=0; i < sun_path_bytes.length; i++)
		{
			addrArr[i + sun_path_index ] = sun_path_bytes[i];
		}


		return addrArr;
	}


	public final static long[] getAddrArrNL( long nl_family, long nl_pad, long nl_pid, long nl_groups, long addrlen )
	{
		//Ǝ֐. sockaddr_nl\̂̑ւȂlongzԂ. 
		//̊֐ł,nl_family\long[]͂߂Ƃϐɑ΂AGfBAϊ͍sȂ.
		//ȂAŌ̗vfaddrleni[.Linuxł͎gp邱Ƃ͂Ȃ낤AÔ߁B

		/*
		struct sockaddr_nl {
			sa_family_t     nl_family;  // AF_NETLINK 
			unsigned short  nl_pad;     // 0 ł 
			pid_t           nl_pid;     // vZX ID 
			__u32           nl_groups;  // }`LXgO[v}XN 
		};
		*/

		long[] addrArr = new long[ 5 ];

		addrArr[0]	= nl_family;
		addrArr[1]	= nl_pad;
		addrArr[2]	= nl_pid;
		addrArr[3]	= nl_groups;
		addrArr[4]	= addrlen;


		return addrArr;
	}



	public final static native long getErrno();		//Ǝ.errno擾.

	public final static native long getCycleCounter();	//Ǝ.TCNJE^(\łΎ擾)



	final private static String BytesToString( byte[] byte_data)
	{
		//NULLŏIbyte[]z񂩂AString쐬.

		long null_value = 0;
		
		int null_position = byte_data.length - 1;
		String str_data = new String("Can't converte");

		
		for (int i=0; i < byte_data.length; i++)
		{
			if ( byte_data[i] == null_value)
			{
				null_position = i;
				break;
			}
		}
		
		byte[] new_byte_data = new byte[ null_position + 1];
		
		System.arraycopy(byte_data, 0, new_byte_data, 0, new_byte_data.length );
		
		try {
			str_data =  new String( new_byte_data, "UTF-8");

		} catch (Exception e) {
			e.printStackTrace();
		}
		
		return str_data;
		
	}


	final private static byte[]	StringToBytes( String str_data )
	{
		byte[] str_bytes = str_data.getBytes();

		byte[] str_bytes_null_terminate = new byte[ str_bytes.length + 1 ];
		System.arraycopy( str_bytes, 0, str_bytes_null_terminate, 0, str_bytes.length );
		str_bytes_null_terminate[ str_bytes.length ] = 0;	//NULL

		return str_bytes_null_terminate;
	}






	/*08/02/02 Append Ԃ[܂ĕKvȂ̂*/

	public final static native int writev(int fd, byte[] jbuffer, int size);
	
	public final static native int uname(String[] utsname);
	
	public final static native long malloc(int size);
	
	public final static native void free(long ptr );
	
	public final static native long munmap(long addr, long len);
	
	public final static native int mprotect(long addr, int len, int prot);
	
	public final static native long mmap(long addr, long len, int prot, int flags, int fd, int pgoff);

	public final static native long getMmapConstValue( String name );
	// 08/11/05 ǉ. PROT_WRITE  MAP_SHARED Ȃǂ̒萔lԂ.


	public final static long stat( String fileName, long[] stat_params)
	{
		// stat64֐ĂяoĊY\̃ô݂.

		/*
		struct stat {
    			dev_t     st_dev;      t@CfoCX ID
    			ino_t     st_ino;      inode ԍ 
    			mode_t    st_mode;     ANZXی 
    			nlink_t   st_nlink;    n[hN̐ 
    			uid_t     st_uid;      L҂̃[U ID 
    			gid_t     st_gid;      L҂̃O[v ID 
    			dev_t     st_rdev;     foCX ID (t@C̏ꍇ) 
    			off_t     st_size;     Ŝ̃TCY (oCgP) 
    			blksize_t st_blksize;  t@CVXe I/O ł̃ubNTCY 
    			blkcnt_t  st_blocks;   蓖ĂꂽubN 
    			time_t    st_atime;    ŏIANZX 
    			time_t    st_mtime;    ŏIC 
    			time_t    st_ctime;    ŏIԕύX 
		};
		*/

		long[] stat64_params = new long[ 20 ];

		long ret = stat64( fileName, stat64_params );

		stat_params[0] = stat64_params[0];
		stat_params[1] = stat64_params[19];
		stat_params[2] = stat64_params[3];
		stat_params[3] = stat64_params[4];
		stat_params[4] = stat64_params[5];
		stat_params[5] = stat64_params[6];
		stat_params[6] = stat64_params[7];
		stat_params[7] = stat64_params[9];
		stat_params[8] = stat64_params[10];
		stat_params[9] = stat64_params[11];
		stat_params[10] = stat64_params[13];
		stat_params[11] = stat64_params[15];
		stat_params[12] = stat64_params[17];

		return ret;
	}
		

	public final static native long stat64( String filename, long[] stat64_params);


	public final static long fstat( long fd, long[] stat_params)
	{
		// fstat64֐ĂяoĊY\̃ô݂.

		/*
		struct stat {
    			dev_t     st_dev;      t@CfoCX ID
    			ino_t     st_ino;      inode ԍ 
    			mode_t    st_mode;     ANZXی 
    			nlink_t   st_nlink;    n[hN̐ 
    			uid_t     st_uid;      L҂̃[U ID 
    			gid_t     st_gid;      L҂̃O[v ID 
    			dev_t     st_rdev;     foCX ID (t@C̏ꍇ) 
    			off_t     st_size;     Ŝ̃TCY (oCgP) 
    			blksize_t st_blksize;  t@CVXe I/O ł̃ubNTCY 
    			blkcnt_t  st_blocks;   蓖ĂꂽubN 
    			time_t    st_atime;    ŏIANZX 
    			time_t    st_mtime;    ŏIC 
    			time_t    st_ctime;    ŏIԕύX 
		};
		*/

		long[] stat64_params = new long[ 20 ];

		long ret = fstat64( fd, stat64_params );

		stat_params[0] = stat64_params[0];
		stat_params[1] = stat64_params[19];
		stat_params[2] = stat64_params[3];
		stat_params[3] = stat64_params[4];
		stat_params[4] = stat64_params[5];
		stat_params[5] = stat64_params[6];
		stat_params[6] = stat64_params[7];
		stat_params[7] = stat64_params[9];
		stat_params[8] = stat64_params[10];
		stat_params[9] = stat64_params[11];
		stat_params[10] = stat64_params[13];
		stat_params[11] = stat64_params[15];
		stat_params[12] = stat64_params[17];

		return ret;
	}



	public final static native long fstat64(long fd, long[] stat64_params);

	public final static native long lstat64( String filename, long[] stat64_params);
	// 09/01/09.




	public final static native long getIoctlConstValue( String contValueName );//Ǝ֐.FIONREADȂǂ̃p[^Ԃ.

	public final static native long ioctl(long fd, long cmd, byte[] argp);

	public final static native long gettimeofday(long[] timeval, long[] timezone);

	public final static native long settimeofday(long[] timeval, long[] timezone);

	


	public final static native long fcntl64(long fd, long cmd, long[] flock);


	public final static native long select(long nfds, long[] readfds, long[] writefds, long[] exceptfds, long[] timeout);




	public final static native long getSignalConstValue( String contValueName );//Ǝ֐.SIG_BLOCkȂǂ̃p[^Ԃ.


	public final static long sigwait( long mask_val, long[] ptr_sig )
	{
		//mask_val̓VOĩXglong^ϐɎ߂̂ł.

		return sigwait( sigmaskToList( mask_val ), ptr_sig );
	}
	public final static native long sigwait(long[] sigmask, long[] ptr_sig);
	// sigmask̓}XNVOii[ꂽzł.
	// jptr_sig͊̃VOiLɂȂԂ߂ɂ,vf1int^z̃CX^Xn.





	public final static long[] sigmaskToList( long sigmask )
	{
		//sigmaskVOïꗗԂ.

		long[] list = null;

		final int sigmask_size = 64;

		long tmp_mask = sigmask;

		Vector<Long> list_v = new Vector();

		for( int i=0; i < sigmask_size; i++)
		{
			if ( (tmp_mask & 0x01) == 1)
			{
				list_v.add( new Long( i ) );
			}

			tmp_mask = tmp_mask >> 1;
		}

		list = new long[ list_v.size() ];
		for( int i=0; i < list.length; i++)
		{
			list[i] = list_v.get(i);
		}


		return list;
	}

	public final static long listToSigmask( long[] list )
	{
		//signalԍi[ꂽXgAsigmaskɑlong^ϐ𐶐.

		long sigmask_val = 0;

		for( int i=0; i < list.length; i++)
		{
			if ( list[i] >= 0 )
			{
				//@̏ꍇ͖炩ɒlsł.
				sigmask_val = sigmask_val | ( 0x01 << list[i] );
			}
		}

		return sigmask_val;
	}


		
	
	public final static long sigsuspend( long mask_val )
	{
		//mask_val̓}XNVOĩXglong^zɎ߂.
		return sigsuspend( sigmaskToList( mask_val ) );
	}
	private final static native long sigsuspend(long[] sigmask);
	
	public final static long sigpause( long mask_val )
	{
		return sigpause( sigmaskToList( mask_val ) );
	}
	private final static native long sigpause(long[] sigmask);




	public final static long sigprocmask(long how, long[] mask, long[][] old_mask)
	{
		//old_mask͗vf1̃_u|C^ł.
		//mask, old_maskɂnullwł


		long[][] tmp_old_mask = new long[1][];

		long ret = sigprocmaskNative( how, mask, tmp_old_mask);
		

		if ( old_mask != null)
		{
			old_mask[0] = tmp_old_mask[0];
		}

		return ret;
	}


	public final static long sigprocmask( long how, Long mask )
	{
		long[] mask_list = sigmaskToList( mask );

		return sigprocmask( how, mask_list, null );
	}


	private final static native long sigprocmaskNative(long how, long[] mask, long[][] old_mask);




	private final static String classPathToNativePath( String class_path )
	{
		String path = new String( class_path );
		
		if ( path.indexOf("class ") == 0)
		{
			path = "L" + path.substring( "class ".length() ) + ";" ;
		}
		else if ( path.indexOf("interface ") == 0)
		{
			path = "L" + path.substring( "interface ".length() ) + ";" ;
		}
		else if ( path.indexOf("[]") >= 0)
		{
			path = path.substring(0, path.indexOf("[]") );
			
			try {
				path = classPathToNativePath( (Class.forName(path) ).toString() );
			} catch (Exception e) {
				e.printStackTrace();
			}

			path = "[" + path ;
		}
		else if ( path.equals("boolean") )
		{
			path= "Z";
		}
		else if ( path.equals("byte") )
		{
			path= "B";
		}
		else if ( path.equals("char") )
		{
			path= "C";
		}
		else if ( path.equals("short") )
		{
			path= "S";
		}
		else if ( path.equals("int") )
		{
			path= "I";
		}
		else if ( path.equals("long") )
		{
			path= "J";
		}
		else if ( path.equals("float") )
		{
			path= "F";
		}
		else if ( path.equals("double") )
		{
			path= "D";
		}
		else if ( path.equals("void") )
		{
			path= "V";
		}


		path = path.replace( "." , "/" );
			


		return path;
	}
		

	public final static String getMethodNativeSignature( Method method)
	{
		//ƂălCeBuR[hĂ΂邱Ƃz肵\bh.
		// MethodCX^XlCeBuŗp\L@̃VOl`Ԃ.

		// VOl`Ɋւ́@http://java.sun.com/j2se/1.5.0/ja/docs/ja/guide/jni/spec/types.html


		posix.print(" posix.getMethodNativeSignature() called\n");

		if ( method == null)
		{
			posix.print(" posix.getMethodNativeSignature() method is null\n");
		}
		else
		{
			posix.print(" posix.getMethodNativeSignature() method is NOT null\n");
			posix.print(" posix.getMethodNativeSignature() method=" + method.toString() + "\n");
		}

		Class[] paramTypes = method.getParameterTypes();


		posix.print(" posix.getMethodNativeSignature() paramTypes=" + paramTypes.toString() + "\n");


		Vector<String> paramTypes_v = new Vector();
		for( int i=0; i < paramTypes.length; i++)
		{
			String path = paramTypes[i].toString();

			path = classPathToNativePath( path );
			
			posix.print("paramTypes[" + i + "]=" + path + "\n");

			paramTypes_v.add( path );
		}

		String param_str = new String();
		for( int i=0; i < paramTypes_v.size(); i++)
		{
			param_str += paramTypes_v.get(i);
		}
		param_str = "(" + param_str + ")";
		
		String ret_path = classPathToNativePath( method.getReturnType().toString() );

		String full_path = param_str + ret_path;

		return full_path;
		
	}


	public final static native long getPthreadConstValue(String  constValueName);


	public final static native long pthreadCreate(long[] pthread_t, long[] pthread_attr_t, Class target_cls, Method routine, Object this_ptr, Object[] args);

	public final static native long pthreadJoin(long pthread_t, Object[] thread_return);

	public final static native void pthreadExit( long retval );	//08/11/05 .


	public final static native long gettid();	//pthread_self()Ɠ`.




	public final static long signal( long signal, Method method, Object this_ptr )
	{
	//this|C^(Ă悢)ƁAsignumvf1long^ẑ݂Ɏvoid^Methodwł.
	//w肵VOiƁAlCeBuR[hJavaVM̃CX^XĂяotNVs.

		SignalHandlerTriger sht = new SignalHandlerTriger();
		sht.setTriger( signal, method, this_ptr );

		return 0;

	}


	//public final static native long sigaction();
		// sigaction()̃eXgp\bh.


	public final static long sigaction( long signum, Class target_cls, Method handler_method, Object this_ptr, Long sa_mask, Long sa_flag)
	{
		posix.print("jnios.posix sigaction()1 called\n");
		return sigaction( signum, target_cls, handler_method, this_ptr, sigmaskToList( sa_mask ), sa_flag );
	}

	public final static long sigaction( long signum, Class target_cls, Method handler_method, Object this_ptr, long[] sa_mask, Long sa_flag)
	{
		//GCAX
		posix.print("jnios.posix sigaction()2 called\n");
		return sigaction( signum, target_cls, handler_method, this_ptr, sa_mask, sa_flag, null );
	
	}
	public final static long sigaction( long signum, Class target_cls, Method handler_method, Object this_ptr, long[] sa_mask, Long sa_flag, Object[] old_act)
	{
		// sigaction()̖{iI. nh̃\bhɎwôvoid^̖߂lƈint^݂̂̃\bhł.
		// ܂ŌObject[]zɂnull,vf5̊mۍςObjectzn.

		posix.print("jnios.posix sigaction()3 called\n");

		Object[] new_act = new Object[ 5 ];

		new_act[0] = target_cls;
		new_act[1] = handler_method;
		new_act[2] = this_ptr;
		new_act[3] = sa_mask;
		new_act[4] = sa_flag;

		if ( old_act == null )
		{
			old_act = new Object[ 5 ];
		}
		
		return sigaction( signum, new_act, old_act );
	}

	public final static native long sigaction( long signum, Object[] new_act, Object[] old_act);



	public final static native long nanosleep( long sec, long nsec, long[] rem );
	// c莞Ԃi[reḿAvf2long^zł.

	public final static native long schedYield();
	// sched_yield()̃bp[.

	public final static native long schedSetAffinity( long pid, boolean[] mask);
	public final static native long schedGetAffinity( long pid, boolean[] mask);
	// w肵vZXsCPUݒ肷. mask͑SRAɂON/OFF̏Ԃtrue/falseŐ擪珇ɐݒ肵.


	// on_cpu_numbersɂONɂ(͂Ă)CPU̔ԍݒ肷.(getł͗]zvfɂ-1ݒ肳)
	public final static long schedSetAffinity( long pid, long[] on_cpu_numbers )
	{
		boolean[] mask = new boolean[ 1024 ];
		for( int i=0; i < mask.length; i++)
		{
			// mask̏
			mask[i] = false;
		}
		for( int i=0; i < on_cpu_numbers.length; i++)
		{
			long cpu_num = on_cpu_numbers[i];
			if (  cpu_num >= 0)
			{
				// w肳ꂽCPU̔ԍ̃}XNZbg.
				mask[ (int)cpu_num ] = true;
			}
		}

		long ret = schedSetAffinity( pid, mask );

		return ret;
	}
	public final static long schedGetAffinity( long pid, long[] on_cpu_numbers )
	{
		boolean[] mask = new boolean[1024];
		for( int i=0; i < mask.length; i++)
		{
			mask[i] = false;
		}
		for( int i=0; i < on_cpu_numbers.length; i++)
		{
			on_cpu_numbers[i] = -1;
		}

		long ret = schedGetAffinity( pid, mask );

		int cpu_num_size = 0;
		for( int i=0; i < mask.length; i++)
		{
			if ( mask[i] == true )
			{
				if ( cpu_num_size < on_cpu_numbers.length )
				{
					on_cpu_numbers[ cpu_num_size ] = i;
				}
			}
		}
		return ret;
	}

	



	public final static native long readNativeMemcpy( byte[] dest_bytes, long native_ptr, long size );
	// lCeBuȃԂJavabyte^zփ]

	public final static native long writeNativeMemcpy( long native_ptr, byte[] src_bytes, long size );
	// Javabyte^z񂩂lCeBuԂփ]



	public final static native long statfs(String path, long[] statfs );
	// }Egꂽt@CEVXeɂĂ̏Ԃ.
	// long^z͈ȉ̍\̂ɏ.
	/*
		struct statfs
		{
		   long    f_type;      t@CVXě^ (LQ) 
		   long    f_bsize;     œKȃubNTCY 
		   long    f_blocks;    t@CVXȇf[^ubN 
 	           long    f_bfree;     t@CVXe̋󂫃ubN 
		   long    f_bavail;    X[p[[Ugp\ȋ󂫃ubN 
		   long    f_files;     t@CVXȇt@Cm[h 
 		   long    f_ffree;     t@CVXe̋󂫃t@Cm[h 
		   fsid_t  f_fsid;      t@CVXe ID 
		   long    f_namelen;   t@C̍ő咷 
		};
	*/
	public final static native long fstatfs( long fd, long[] statfs);
	// fBXNv^[ fd ɂĎQƂI[vꂽt@Cɂ statfs()ƓԂ.

	// public final static native long getStatfsFTypeValue(String  constValueName);
	// f_typẽlCeBu̒萔lԂ.
	// Cygwinmagic.hɑ`t@C݂Ȃ߁Agp͕s.


	// public final static long getdents(long fd, long[] d_ino_arr, long[] d_off_arr, long[] d_reclean_arr, String[] d_name_arr, long d_type_arr, long count );
	// getdents()ɑ鏈s. dirent\̃obt@̑ɍ\̗̓vfʌ̔zƂĈ.
	// getdents()̂LinuxƎ̃VXeR[Ȃ̂ŁAreaddir()bvĎ.
	// I[vς݂̃fBXNv^fBNg|C^擾ɂ, fdopendir()gp.
	// openat


	public final static native long opendir( String name );
	// fBNg name ɑΉfBNgXg[I[vÃXg[ւ̃|C^Ԃ

	public final static native Object[] readdir( long dirp );
	// dirp wfBNgXg[̒ŁÃfBNgGg\ dirent \̂ɏObject^zԂ.
	// zvf { Long d_ino, Long d_off, Long d_reclen, Long d_type, String d_name }; ł.
	// G[邩AfBNgXg[̖ɒBꍇA null Ԃ.

	public final static native long closedir( long dirp );
	// dirp Ɋ֘AtꂽfBNgXg[N[Y.
	


	public final static native long poll( long[] fds, String[][] events, String[][] revents, long timeoutMillis );
	// 09/06/17 .
	// fdsz̗vf́Aeventsreventsw肷.
	// events,reventśAString^ }XN/Lb`ꂽ Cxgi[zł.


	public final static long waitpid( long pid, long[] status, String[] options )
	{
		// 09/06/17 .
		// status͗vfPlong^z.
		// options͎w肵IvV̕i[z.
		// {\bhwaitpidT()̃GCAXł.

		long options_val = 0;

		for( int i=0; i < options.length; i++)
		{
			long constVal = getWaitConstValue( options[i] );
			if ( constVal >= 0 )
			{
				options_val |= constVal;
			}
		}
		long ret = waitpidT( pid, status, options_val );
		return ret;
	}

	public final static native long waitpidT( long pid, long[] status, long options);
	//waitpid()Ăяo̎.

	public final static native long getWaitConstValue( String constValueName );
	// waitpid()ŗpIvVȂǂ̒萔lԂ.



	public final static native long getitimer(long which, long[] curr_value);
	// 10/01/27. 

	public final static native long setitimer(long which, long[] new_value, long[] old_value );
	// 10/01/27. 


	public final static native long getrusage(long who, long[] rusage_arr );
	// 10/01/27. 

	

	



}








class SignalHandlerTriger extends Thread
{
	//signal(),sigaction()ŌĂяonh\bh̋NpNX.
	//XbhŔCӂ̃VOisꂽۂɁACӂ̃\bhĂяo.
	// Method̈Long^̃CX^Xł.

	long signal = 0;
	Method method = null;
	Object this_ptr = null;


	public void setTriger( long signal, Method method, Object this_ptr )
	{
		this.signal = signal;
		this.method = method;
		this.this_ptr = this_ptr;

		this.start();
	}

	public void run()
	{
		//publicO璼ڌĂł͂ȂȂ.

		Long sig_num_ptr = signal;

	
		// http://d.hatena.ne.jp/yupo5656/20060114/p1
		// "sigwaitŏVOi()A炩sigprocmaskŃubNĂ`ɂȂ܂"Ƃ̂.

		long sigprocmask_ret = posix.sigprocmask( posix.getSignalConstValue("SIG_BLOCK") /*SIG_BLOCK*/, signal );

		posix.print("SignalHandlerTriger.run() signal=" + signal + " SIG_BLOCK=" + posix.getSignalConstValue("SIG_BLOCK") + " sigprocmask_ret = " + sigprocmask_ret + "\n");

		posix.print("SignalHandlerTriger.run() sigwait() call\n");

		long[] sig_num = new long[1];
		posix.sigwait( signal, sig_num);

		posix.print("SignalHandlerTriger.run() sigwait() returned\n");


		//SIG_BLOCK.
		posix.sigprocmask( posix.getSignalConstValue("SIG_UNBLOCK"), signal );

		Object[] method_params = { sig_num_ptr };
		
		posix.print("SignalHandlerTriger.run() invoke() call\n");

		invoke( this_ptr, method, method_params );

		posix.print("SignalHandlerTriger.run() invoke() returned\n");
	}

	private void invoke(Object instance, Method method, Object[] args)
	{

		//ɑ݂CX^X𗘗pthis|C^ēn.
		//static\bh̏ꍇł,Ȃ.

		Object ret = null;
		
		try{
			ret = method.invoke( instance, (Object[])args);
		} catch (Exception e ) {
			e.printStackTrace();
		}

		return ;
	}

}













