/*
 * Copyright 1999 Finn Bock.
 * Permission to use, copy and distribute this software is hereby granted, 
 * provided that the above copyright notice appear in all copies and that 
 * both that copyright notice and this permission notice appear.
 * 
 * No Warranty
 * The software is provided "as is" without warranty of any kind.
 * 
 * If you have questions regarding this software, contact:
 *    Finn Bock, bckfnn@pipmail.dknet.dk
 * 
 * This program contains material copyrighted by:
 * Copyright 1991-1995 by Stichting Mathematisch Centrum, Amsterdam,
 * The Netherlands.
 *
 */

#include "posixmodule.h"

#include "pyutil.h"
#include "jnios_PosixFile.h"

#define _PACKAGE(f) Java_jnios_PosixFile_##f


#include <sys/types.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef MS_WIN32
#define ftruncate _chsize
#define fileno _fileno
#define HAVE_FTRUNCATE
#endif

#ifdef macintosh
#ifdef USE_GUSI
#define HAVE_FTRUNCATE
#endif
#endif

#ifdef THINK_C
#define HAVE_FOPENRF
#endif
#ifdef __MWERKS__
/* Mwerks fopen() doesn't always set errno */
#define NO_FOPEN_ERRNO
#endif

#include <errno.h>
#include <string.h>



static int isClosed(JNIEnv* env, FILE* fp) {
    if (fp == NULL) {
	throwPyException(env, excType(env, "ValueError"), "I/O operation on closed file");
    }
    return fp == NULL;
}



jclass *PosixFileClass;





typedef int (*fClose)(FILE *);

JNIEXPORT jint JNICALL _PACKAGE(close)(JNIEnv *env, jobject jobj) 
{
    int sts = 0;
    FILE* f_fp = (FILE*)getLongInstanceField(env, PosixFileClass, jobj, "f_fp");

    if (f_fp != NULL) {
	//int (*f_close)(FILE *);
	fClose f_close = (fClose)getLongInstanceField(env, PosixFileClass, jobj, "f_close");
	if (f_close != NULL) {
	    errno = 0;
	    sts = (*f_close)(f_fp);
	}
	setLongInstanceField(env, PosixFileClass, jobj, "f_fp", 0);
    }
    if (sts == EOF)
	throwFromErrno(env, excType(env, "IOError"));
    if (sts != 0)
	return sts;
    return 0; 
}


JNIEXPORT jint JNICALL _PACKAGE(fileno)(JNIEnv *env, jobject jobj)
{
    FILE* f_fp = (FILE*)getLongInstanceField(env, PosixFileClass, jobj, "f_fp");
    if (isClosed(env, f_fp))
	return 0;
    return fileno(f_fp);
}

JNIEXPORT void JNICALL _PACKAGE(flush)(JNIEnv *env, jobject jobj)
{
    FILE* f_fp = (FILE*)getLongInstanceField(env, PosixFileClass, jobj, "f_fp");
    if (isClosed(env, f_fp))
	return;
	
    errno = 0;
    int res = fflush(f_fp);
    if (res != 0) {
	throwFromErrno(env, excType(env, "IOError"));
	clearerr(f_fp);
    }
}

JNIEXPORT jobject JNICALL _PACKAGE(tell)(JNIEnv *env, jobject jobj)
{
    FILE* f_fp = (FILE*)getLongInstanceField(env, PosixFileClass, jobj, "f_fp");
    if (isClosed(env, f_fp))
	return NULL;
    errno = 0;
#if defined(HAVE_FTELLO) && defined(HAVE_LARGEFILE_SUPPORT)
    off_t offset = ftello(f_fp);
#elif defined(HAVE_FTELL64) && defined(HAVE_LARGEFILE_SUPPORT)
    off_t offset = ftell64(f_fp);
#else
    off_t offset = ftell(f_fp);
#endif
    if (offset == -1) {
	throwFromErrno(env, excType(env, "IOError"));
	clearerr(f_fp);
	return NULL;
    }
#if !defined(HAVE_LARGEFILE_SUPPORT)
    return NewPyInteger(env, offset);
#else
    return NewPyLong(env, offset);
#endif
}



#ifdef HAVE_FTRUNCATE
JNIEXPORT void JNICALL _PACKAGE(truncate)(JNIEnv *env, jobject jobj,
    jlong newsize)
{
    FILE* f_fp = (FILE*)getLongInstanceField(env, PosixFileClass, jobj, "f_fp");
    if (isClosed(env, f_fp))
	return;

    if (newsize == -1L) {
	/* Default to current position*/
	errno = 0;
#if defined(HAVE_FTELLO) && defined(HAVE_LARGEFILE_SUPPORT)
	newsize =  ftello(f_fp);
#elif defined(HAVE_FTELL64) && defined(HAVE_LARGEFILE_SUPPORT)
	newsize =  ftell64(f_fp);
#else
	newsize =  ftell(f_fp);
#endif
	if (newsize == -1) {
	    throwFromErrno(env, excType(env, "IOError"));
	    clearerr(f_fp);
	    return;
	}
    }

    errno = 0;
    int ret = fflush(f_fp);
    if (ret == 0) {
	errno = 0;
	ret = ftruncate(fileno(f_fp), newsize);
    }
    if (ret != 0) {
	throwFromErrno(env, excType(env, "IOError"));
	clearerr(f_fp);
	return;
    }
}
#endif /* HAVE_FTRUNCATE */





JNIEXPORT jint JNICALL _PACKAGE(isatty)(JNIEnv *env, jobject jobj)
{
    FILE* f_fp = (FILE*)getLongInstanceField(env, PosixFileClass, jobj, "f_fp");
    if (isClosed(env, f_fp))
	return 0;
    return isatty((int)fileno(f_fp));
}


#ifndef DONT_HAVE_FSTAT
#define HAVE_FSTAT

#include <sys/types.h>
#include <sys/stat.h>

#endif

#if BUFSIZ < 8192
#define SMALLCHUNK 8192
#else
#define SMALLCHUNK BUFSIZ
#endif

#if SIZEOF_INT < 4
#define BIGCHUNK  (512 * 32)
#else
#define BIGCHUNK  (512 * 1024)
#endif

static size_t
new_buffersize(FILE* f_fp, size_t currentsize)
{
#ifdef HAVE_FSTAT
    long pos, end;
    struct stat st;
    if (fstat(fileno(f_fp), &st) == 0) {
	end = st.st_size;
	/* The following is not a bug: we really need to call lseek()
	   *and* ftell().  The reason is that some stdio libraries
	   mistakenly flush their buffer when ftell() is called and
	   the lseek() call it makes fails, thereby throwing away
	   data that cannot be recovered in any way.  To avoid this,
	   we first test lseek(), and only call ftell() if lseek()
	   works.  We can't use the lseek() value either, because we
	   need to take the amount of buffered data into account.
	   (Yet another reason why stdio stinks. :-) */
	pos = lseek(fileno(f_fp), 0L, SEEK_CUR);
	if (pos >= 0)
	    pos = ftell(f_fp);
	if (pos < 0)
	    clearerr(f_fp);
	if (end > pos && pos >= 0)
	    return currentsize + end - pos + 1;
	/* Add 1 so if the file were to grow we'd notice. */
    }
#endif
    if (currentsize > SMALLCHUNK) {
	/* Keep doubling until we reach BIGCHUNK;
	   then keep adding BIGCHUNK. */
	if (currentsize <= BIGCHUNK)
	    return currentsize + currentsize;
	else
	    return currentsize + BIGCHUNK;
    }
    return currentsize + SMALLCHUNK;
}


JNIEXPORT jstring JNICALL _PACKAGE(read)(JNIEnv *env, jobject jobj,
    jlong bytesrequested)
{
    FILE* f_fp = (FILE*)getLongInstanceField(env, PosixFileClass, jobj, "f_fp");
    if (isClosed(env, f_fp))
	return NULL;

    size_t bytesread, buffersize, chunksize;
	
    if (bytesrequested < 0)
	buffersize = new_buffersize(f_fp, 0);
    else
	buffersize = bytesrequested;

    char* v = (char*)malloc(buffersize);
    if (v == NULL)
	return NULL;
    bytesread = 0;
    for (;;) {
	errno = 0;
	chunksize = fread(v + bytesread, 1,
				  buffersize - bytesread, f_fp);
	if (chunksize == 0) {
	    if (!ferror(f_fp))
		break;
	    throwFromErrno(env, excType(env, "IOError"));
	    clearerr(f_fp);
	    free(v);
	    return NULL;
	}
	bytesread += chunksize;
	if (bytesread < buffersize)
	    break;
	if (bytesrequested < 0) {
	    buffersize = new_buffersize(f_fp, buffersize);
	    v = (char *)realloc(v, buffersize);
	    if (v == NULL) {
		throwPyException(env, excType(env, "MemoryError"), "realloc failed");
	    	return NULL;
	    }
	}
    }
    jstring jret = NewString(env, v, bytesread);
    free(v);
    return jret;
}


static jstring getline(JNIEnv *env, FILE* fp, int n)
{
    register int c;

    int n2 = n > 0 ? n : 100;
    char* v = (char*) malloc(n2);
    if (v == NULL)
	return NULL;
    char* buf = v;
    char* end = buf + n2;

    for (;;) {
	if ((c = getc(fp)) == EOF) {
	    clearerr(fp);
	    //if (PyErr_CheckSignals()) {
	    //	Py_DECREF(v);
	    //	return NULL;
	    //}
	    if (n < 0 && buf == v) {
		throwPyException(env, excType(env, "EOFError"), "EOF when reading a line");
		free(v);
		return NULL;
	    }
	    break;
	}
	if ((*buf++ = c) == '\n') {
	    if (n < 0)
		buf--;
	    break;
	}
	if (buf == end) {
	    if (n > 0)
		break;
	    int n1 = n2;
	    n2 += 1000;
	    v = (char *) realloc(v, n2);
	    if (v == NULL) {
		throwPyException(env, excType(env, "MemoryError"), "realloc failed");
		return NULL;
	    }
	    buf = v + n1;
	    end = v + n2;
	}
    }

    jstring jret = NewString(env, v, buf - v);
    free(v);
    return jret;
}


JNIEXPORT jstring JNICALL _PACKAGE(readline)(JNIEnv *env, jobject jobj, 
    jint jn)
{
    if (jn == 0)
	return env->NewStringUTF("");

    FILE* f_fp = (FILE*)getLongInstanceField(env, PosixFileClass, jobj, "f_fp");
    if (isClosed(env, f_fp))
	return NULL;

    if (jn < 0)
	jn = 0;
    return getline(env, f_fp, jn);
}



JNIEXPORT jobject JNICALL _PACKAGE(readlines)(JNIEnv *env, jobject jobj,
    jint jsizehint)
{
    char small_buffer[SMALLCHUNK];
    char *buffer = small_buffer;
    size_t buffersize = SMALLCHUNK;

    char *big_buffer = NULL;
    size_t nfilled = 0;
    size_t nread;
    size_t totalread = 0;
	char *p, *q, *end;
	int err;

    FILE* f_fp = (FILE*)getLongInstanceField(env, PosixFileClass, jobj, "f_fp");
    if (isClosed(env, f_fp))
	return NULL;

    jobject list = NewPyList(env);
    CHECK(list, return NULL, "NewPyList failed");

    for (;;) {
	errno = 0;
	nread = fread(buffer+nfilled, 1, buffersize-nfilled, f_fp);
	if (nread == 0) {
	    jsizehint = 0;
	    if (!ferror(f_fp))
		break;
	    throwFromErrno(env, excType(env, "IOError"));
	    clearerr(f_fp);
	error:
	    list = NULL;
	    goto cleanup;
	}
	totalread += nread;
	p = (char*)memchr(buffer+nfilled, '\n', nread);
	if (p == NULL) {
	    /* Need a larger buffer to fit this line */
	    nfilled += nread;
	    buffersize *= 2;
	    if (big_buffer == NULL) {
		/* Create the big buffer */
		buffer = big_buffer = (char*) malloc(buffersize);
		if (big_buffer == NULL)
		    goto error;
		memcpy(big_buffer, small_buffer, nfilled);
	    }
	    else {
		/* Grow the big buffer */
		buffer = big_buffer = (char *)realloc(big_buffer, buffersize);
	    }
	    continue;
	}
	end = buffer+nfilled+nread;
	q = buffer;
	do {
	    /* Process complete lines */
	    p++;
	    jstring jline = NewString(env, q, p-q);
	    if (jline == NULL)
		goto error;
	    jobject line = NewPyString(env, jline);
	    if (line == NULL)
		goto error;
	    err = PyObject_invoke(env, list, "append", line);
	    if (err != 0)
		goto error;
	    q = p;
	    p = (char*)memchr(q, '\n', end-q);
	} while (p != NULL);
	/* Move the remaining incomplete line to the start */
	nfilled = end-q;
	memmove(buffer, q, nfilled);
	if (jsizehint > 0)
	    if (totalread >= (size_t)jsizehint)
		break;
    }
    if (nfilled != 0) {
	/* Partial last line */
	jstring line = NewString(env, buffer, nfilled);
	if (line == NULL)
	    goto error;
	if (jsizehint > 0) {
	    /* Need to complete the last line */
	    jstring rest = getline(env, f_fp, 0);
	    if (rest == NULL) {
		goto error;
	    }
	    line = StringConcat(env, line, rest);
	    if (line == NULL)
		goto error;
	}
	err = PyObject_invoke(env, list, "append", line);
	if (err != 0)
	    goto error;
    }
  cleanup:
    if (big_buffer) {
	free(big_buffer);
    }
    return list;
}



JNIEXPORT void JNICALL _PACKAGE(seek)(JNIEnv *env, jobject jobj,
    jlong joffset, jint jwhence)
{
    FILE* f_fp = (FILE*)getLongInstanceField(env, PosixFileClass, jobj, "f_fp");
    if (isClosed(env, f_fp))
	return;
	
    errno = 0;
#if defined(HAVE_FSEEKO)
    int ret = fseeko(f_fp, joffset, jwhence);
#elif defined(HAVE_FSEEK64)
    int ret = fseek64(f_fp, joffset, jwhence);
#else
    int ret = fseek(f_fp, joffset, jwhence);
#endif
    if (ret != 0) {
	throwFromErrno(env, excType(env, "IOError"));
	clearerr(f_fp);
    }
    return;
}


JNIEXPORT void JNICALL _PACKAGE(setbufsize)(JNIEnv *env, jobject jobj,
    jint jbufsize)
{
    if (jbufsize >= 0) {
	FILE* f_fp = (FILE*)getLongInstanceField(env, PosixFileClass, jobj, "f_fp");
#ifdef HAVE_SETVBUF
	int type;
	switch (jbufsize) {
	case 0:
	    type = _IONBF;
	    break;
	case 1:
	    type = _IOLBF;
	    jbufsize = BUFSIZ;
	    break;
	default:
	    type = _IOFBF;
	}
	setvbuf(f_fp, (char *)NULL, type, jbufsize);
#else /* !HAVE_SETVBUF */
	if (bufsize <= 1)
	    setbuf(f_fp, (char *)NULL);
#endif /* !HAVE_SETVBUF */
    }
}


JNIEXPORT void JNICALL _PACKAGE(write)(JNIEnv *env, jobject jobj,
    jstring jline)
{
    FILE* f_fp = (FILE*)getLongInstanceField(env, PosixFileClass, jobj, "f_fp");
    if (isClosed(env, f_fp))
	return;

    setIntInstanceField(env, PosixFileClass, jobj, "softspace", 0);
    int n = env->GetStringLength(jline);

    errno = 0;
    int n2 = fwrite(StringAsChars(env, jline), 1, n, f_fp);
    if (n2 != n) {
	throwFromErrno(env, excType(env, "IOError"));
	clearerr(f_fp);
    }
}



JNIEXPORT void JNICALL _PACKAGE(writelines)(JNIEnv *env, jobject jobj, 
    jobjectArray jlines) 
{
    FILE* f_fp = (FILE*)getLongInstanceField(env, PosixFileClass, jobj, "f_fp");
    if (isClosed(env, f_fp))
	return;

    int n = env->GetArrayLength(jlines);

    setIntInstanceField(env, PosixFileClass, jobj, "softspace", 0);

    errno = 0;
    for (int i = 0; i < n; i++) {
	jstring line = (jstring)env->GetObjectArrayElement(jlines, i);
	int len = env->GetStringLength(line);
	int nwritten = fwrite(StringAsChars(env, line), 1, len, f_fp);
	if (nwritten != len) {
	    throwFromErrno(env, excType(env, "IOError"));
	    clearerr(f_fp);
	    return;
	}
    }
}


